"""Configurable values exposed to user as environment variables"""
import logging
import os

# Changes the log level
LOG_LEVEL: str = os.getenv("LOG_LEVEL", "ERROR")

logging.basicConfig(level=LOG_LEVEL)

# Enables verbose logging.  It is recommended to also use ``LOG_LEVEL=DEBUG``.
VERBOSE_LOGS: bool = False if os.getenv("VERBOSE_LOGS", "FALSE") == "FALSE" else True

# https://gdal.org/user/virtual_file_systems.html#vsicurl-http-https-ftp-files-random-access
# Defines the number of bytes read in the first GET request at file opening
# Can help performance when reading images with a large header
INGESTED_BYTES_AT_OPEN: int = os.getenv("INGESTED_BYTES_AT_OPEN", 16384)

# https://trac.osgeo.org/gdal/wiki/ConfigOptions#VSI_CACHE
# Determines if in-memory block caching is enabled
ENABLE_CACHE: bool = True if os.getenv(
    "ENABLE_CACHE", "TRUE"
).upper() == "TRUE" else False

# https://trac.osgeo.org/gdal/wiki/ConfigOptions#GDAL_HTTP_MERGE_CONSECUTIVE_RANGES
# Determines if consecutive range requests are merged into a single request, reducing the number of HTTP GET range
# requests required to read consecutive internal image tiles
HTTP_MERGE_CONSECUTIVE_RANGES: bool = True if os.getenv(
    "HTTP_MERGE_CONSECUTIVE_RANGES", "FALSE"
).upper() == "TRUE" else False


# Determines if internal tiles outside the bounds of the IFD are read. When set to ``TRUE`` (default), if a partial read
# isn't fully covered by internal tiles, missing tiles will be created using the fill value defined by the
# ``BOUNDLESS_READ_FILL_VALUE`` config option. When set to ``FALSE``, an exception will be raised instead
BOUNDLESS_READ: bool = False if os.getenv(
    "BOUNDLESS_READ", "TRUE"
) == "FALSE" else True


# Determines the fill value used for boundless reads
BOUNDLESS_READ_FILL_VALUE: int = int(os.getenv("BOUNDLESS_READ_FILL_VALUE", "0"))



"""COG mixins for partial reads"""
import asyncio
import abc
from dataclasses import dataclass
import math
from typing import List, Optional, Tuple, Union

import affine
import numpy as np
from PIL import Image

from .ifd import ImageIFD, MaskIFD
from .utils import run_in_background


NpArrayType = Union[np.ndarray, np.ma.masked_array]


@dataclass
class TileMetadata:
    # top left corner of the partial read
    tlx: float
    tly: float
    # width and height of the partial read (# of pixels)
    width: int
    height: int
    # width and height of each block (# of pixels)
    tile_width: int
    tile_height: int
    # range of internal x/y blocks which intersect the partial read
    xmin: int
    ymin: int
    xmax: int
    ymax: int
    # expected number of bands
    bands: int
    # numpy data type
    dtype: np.dtype
    # overview level (where 0 is source)
    ovr_level: int


@dataclass
class PartialReadBase(abc.ABC):
    @property
    @abc.abstractmethod
    def is_masked(self) -> bool:
        """Check if the image has an internal mask"""
        ...

    @property
    @abc.abstractmethod
    def nodata(self) -> Optional[int]:
        """Return nodata value or None if there is not a nodata value"""
        ...

    @property
    @abc.abstractmethod
    def overviews(self) -> List[int]:
        """Return decimation factor for each overview (2**zoom)"""
        ...

    @abc.abstractmethod
    def geotransform(self, ovr_level: int = 0) -> affine.Affine:
        """Return the geotransform of the image at a specific overview level (defaults to native resolution)"""
        ...

    @abc.abstractmethod
    async def get_tile(self, x: int, y: int, z: int) -> NpArrayType:
        """Request an internal image tile at the specified row (x), column (y), and overview (z)"""
        ...

    @abc.abstractmethod
    async def read(
        self, bounds: Tuple[float, float, float, float], shape: Tuple[int, int], resample_method: int
    ) -> Union[np.ndarray, np.ma.masked_array]:
        """Do a partial read"""
        ...

    @property
    def _add_mask(self) -> bool:
        """Determine if a mask needs to be added to the array"""
        if self.is_masked:
            return True
        if self.nodata is not None:
            return True
        return False


    @staticmethod
    def _intersect_bounds(
        read_bounds: Tuple[float, float, float, float],
        cog_bounds: Tuple[float, float, float, float]
    ) -> bool:
        """
        Determine if a bounding box intersects another bounding box

        https://github.com/cogeotiff/rio-tiler/blob/2.0a11/rio_tiler/utils.py#L254-L283
        """
        return (
            (cog_bounds[0] < read_bounds[2])
            and (cog_bounds[2] > read_bounds[0])
            and (cog_bounds[3] > read_bounds[1])
            and (cog_bounds[1] < read_bounds[3])
        )

    def _get_overview_level(
        self, bounds: Tuple[float, float, float, float], width: int, height: int
    ) -> int:
        """
        Calculate appropriate overview level given request bounds and shape (width + height).  Based on rio-tiler:
        https://github.com/cogeotiff/rio-tiler/blob/v2/rio_tiler/utils.py#L79-L135
        """
        src_res = self.geotransform().a
        target_gt = affine.Affine.translation(
            bounds[0], bounds[3]
        ) * affine.Affine.scale(
            (bounds[2] - bounds[0]) / width, (bounds[1] - bounds[3]) / height
        )
        target_res = target_gt.a

        ovr_level = 0
        if target_res > src_res:
            # Decimated resolution at each overview
            overviews = [src_res * decim for decim in self.overviews]
            for ovr_level in range(ovr_level, len(overviews) - 1):
                ovr_res = src_res if ovr_level == 0 else overviews[ovr_level]
                if (ovr_res < target_res) and (overviews[ovr_level + 1] > target_res):
                    break
                if abs(ovr_res - target_res) < 1e-1:
                    break
            else:
                ovr_level = len(overviews) - 1

        return ovr_level

    def _calculate_image_tiles(
        self,
        bounds: Tuple[float, float, float, float],
        tile_width: int,
        tile_height: int,
        band_count: int,
        ovr_level: int,
        dtype: np.dtype,
    ) -> TileMetadata:
        """
        Internal method to calculate which images tiles need to be requested for a partial read.  Also returns all of
        the required metadata about the image tiles to perform a partial read
        """
        geotransform = self.geotransform(ovr_level)
        invgt = ~geotransform

        # Project request bounds to pixel coordinates relative to geotransform of the overview
        tlx, tly = invgt * (bounds[0], bounds[3])
        brx, bry = invgt * (bounds[2], bounds[1])

        # Calculate tiles
        xmin = math.floor((tlx + 1e-6) / tile_width)
        xmax = math.floor((brx + 1e-6) / tile_width)
        ymax = math.floor((bry + 1e-6) / tile_height)
        ymin = math.floor((tly + 1e-6) / tile_height)

        tile_bounds = (
            xmin * tile_width,
            ymin * tile_height,
            (xmax + 1) * tile_width,
            (ymax + 1) * tile_height,
        )

        # Create geotransform for the fused image
        _tlx, _tly = geotransform * (tile_bounds[0], tile_bounds[1])
        fused_gt = affine.Affine(
            geotransform.a, geotransform.b, _tlx, geotransform.d, geotransform.e, _tly
        )
        inv_fused_gt = ~fused_gt
        xorigin, yorigin = [round(v) for v in inv_fused_gt * (bounds[0], bounds[3])]

        return TileMetadata(
            tlx=xorigin,
            tly=yorigin,
            width=round(brx - tlx),
            height=round(bry - tly),
            xmin=xmin,
            ymin=ymin,
            xmax=xmax,
            ymax=ymax,
            tile_width=tile_width,
            tile_height=tile_height,
            bands=band_count,
            dtype=dtype,
            ovr_level=ovr_level,
        )

    def _init_array(self, img_tiles: TileMetadata) -> NpArrayType:
        """
        Initialize an empty numpy array with the same shape of the partial read.  Individual blocks are mosaiced into
        this array as they are requested
        """
        fused = np.zeros(
            (
                img_tiles.bands,
                (img_tiles.ymax + 1 - img_tiles.ymin) * img_tiles.tile_height,
                (img_tiles.xmax + 1 - img_tiles.xmin) * img_tiles.tile_width,
            )
        ).astype(img_tiles.dtype)
        if self._add_mask:
            fused = np.ma.masked_array(fused)
        return fused


    @staticmethod
    def _stitch_image_tile(
        arr: NpArrayType,
        fused_arr: NpArrayType,
        idx: int,
        idy: int,
        tile_width: int,
        tile_height: int,
    ) -> None:
        """Mosaic an array into a larger array"""
        fused_arr[
            :,
            idy * tile_height : (idy + 1) * tile_height,
            idx * tile_width : (idx + 1) * tile_width,
        ] = arr
        if np.ma.is_masked(arr):
            fused_arr.mask[
                :,
                idy * tile_height : (idy + 1) * tile_height,
                idx * tile_width : (idx + 1) * tile_width,
            ] = arr.mask


    async def _get_and_stitch_tile(
        self,
        xtile: int,
        ytile: int,
        idx: int,
        idy: int,
        img_tiles: TileMetadata,
        fused_arr: NpArrayType
    ) -> None:
        """Asynchronously request an internal tile and stitch into an array"""
        tile = await self.get_tile(xtile, ytile, img_tiles.ovr_level)
        self._stitch_image_tile(tile, fused_arr, idx, idy, img_tiles.tile_width, img_tiles.tile_height)


    async def _request_tiles(self, img_tiles: TileMetadata) -> NpArrayType:
        """Concurrently request the image tiles and mosaic into a larger array"""
        img_arr = self._init_array(img_tiles)
        tile_tasks = []
        for idx, xtile in enumerate(range(img_tiles.xmin, img_tiles.xmax + 1)):
            for idy, ytile in enumerate(range(img_tiles.ymin, img_tiles.ymax + 1)):
                get_tile_task = asyncio.create_task(
                    self._get_and_stitch_tile(xtile, ytile, idx, idy, img_tiles, img_arr)
                )
                tile_tasks.append(get_tile_task)
        await asyncio.gather(*tile_tasks)
        return img_arr

    def _clip_array(self, arr: NpArrayType, img_tiles: TileMetadata) -> NpArrayType:
        """Clip a numpy array to the extent of the parial read via slicing"""
        return arr[
            :,
            img_tiles.tly : img_tiles.tly + img_tiles.height,
            img_tiles.tlx : img_tiles.tlx + img_tiles.width,
        ]

    def _resample(
        self, clipped: NpArrayType, img_tiles: TileMetadata, out_shape: Tuple[int, int], resample_method: int
    ) -> NpArrayType:
        """Resample a numpy array to the desired shape"""
        img = Image.fromarray(np.rollaxis(clipped, 0, 3))
        resized = img.resize((out_shape[0], out_shape[1]), resample=Image.BILINEAR)
        resized = np.rollaxis(np.array(resized), 2, 0)
        if self._add_mask:
            mask = Image.fromarray(clipped.mask[0,...])
            resized_mask = np.array(mask.resize((out_shape[0], out_shape[1]), resample=resample_method))
            resized_mask = np.stack((resized_mask for _ in range(img_tiles.bands)))
            resized = np.ma.masked_array(resized, resized_mask)
        return resized

    def _postprocess(
        self, arr: NpArrayType, img_tiles: TileMetadata, out_shape: Tuple[int, int], resample_method: int
    ) -> NpArrayType:
        """Wrapper around ``_clip_array`` and ``_resample`` to postprocess the partial read"""
        return self._resample(
            self._clip_array(arr, img_tiles), img_tiles=img_tiles, out_shape=out_shape, resample_method=resample_method
        )


@dataclass
class PartialReadInterface(PartialReadBase):

    @staticmethod
    def _extract_tile(
        ifd: Union[ImageIFD, MaskIFD], img_bytes: bytes, tile_index: int, offset: int
    ) -> bytes:
        """Extract a tile from the merged range request"""
        byte_count = ifd.TileByteCounts[tile_index]
        tile_start = ifd.TileOffsets[tile_index] - offset
        tile_bytes = img_bytes[tile_start : tile_start + byte_count]
        return tile_bytes

    @staticmethod
    def _merge_range_requests(
        ifd: Union[ImageIFD, MaskIFD], tile_indices: List[int], offset: int
    ) -> Tuple[int, int]:
        """Determine offset and byte count for a range request across given tile indices"""
        byte_count = (
            ifd.TileOffsets[max(tile_indices)] + ifd.TileByteCounts[max(tile_indices)]
        )
        return (offset, byte_count - offset - 1)

    async def _request_merged_tile(
        self, arr: NpArrayType, indices: List[Tuple[int, int, int]], img_tiles: TileMetadata
    ) -> None:
        """Request a range, extract/decompress/mosaic each tile"""
        tile_indices = [idx[0] for idx in indices]
        # Request image data
        futures = []
        ifd = self.ifds[img_tiles.ovr_level]
        offset = ifd.TileOffsets[min(tile_indices)]
        tile_task = asyncio.create_task(
            self._file_reader.range_request(
                *self._merge_range_requests(ifd, tile_indices, offset)
            )
        )
        futures.append(tile_task)

        if self._add_mask:
            # Request mask data
            mask_ifd = self.mask_ifds[img_tiles.ovr_level]
            mask_offset = mask_ifd.TileOffsets[min(tile_indices)]
            mask_task = asyncio.create_task(
                self._file_reader.range_request(
                    *self._merge_range_requests(mask_ifd, tile_indices, mask_offset)
                )
            )
            futures.append(mask_task)

        response = await asyncio.gather(*futures)

        # Compression is applied to each block, so we need to extract and decompress each tile in the merged request
        # TODO: Parallelize
        for (tile_idx, idx, idy) in indices:
            # Extract the tile
            tile_bytes = self._extract_tile(ifd, response[0], tile_idx, offset)
            # Decompress the tile
            decoded = await run_in_background(ifd._decompress, tile_bytes)
            if self._add_mask:
                # Extract mask
                mask_ifd = self.mask_ifds[img_tiles.ovr_level]
                mask_bytes = self._extract_tile(
                    mask_ifd, response[1], tile_idx, mask_offset
                )
                # Decompress and apply mask
                mask_decoded = await run_in_background(ifd._decompress_mask, mask_bytes)
                decoded = np.ma.masked_array(
                    decoded, np.invert(np.broadcast_to(mask_decoded, decoded.shape))
                )
            # Mosaic
            self._stitch_image_tile(
                decoded, arr, idx, idy, img_tiles.tile_width, img_tiles.tile_height
            )

    async def _request_merged_tiles(self, img_tiles: TileMetadata) -> NpArrayType:
        """Do a partial read with merged range requests"""
        futures = []
        ifd = self.ifds[img_tiles.ovr_level]
        # Create the array
        img_arr = self._init_array(img_tiles)
        for idy, ytile in enumerate(range(img_tiles.ymin, img_tiles.ymax + 1)):
            # Merge requests across rows
            indices = []
            for idx, xtile in enumerate(range(img_tiles.xmin, img_tiles.xmax + 1)):
                tile_index = (ytile * ifd.tile_count[0]) + xtile
                indices.append((tile_index, idx, idy))
            # Do the request
            merged_tile_task = asyncio.create_task(
                self._request_merged_tile(img_arr, indices, img_tiles)
            )
            futures.append(merged_tile_task)
        await asyncio.gather(*futures)
        return img_arr



import asyncio
from functools import partial
from typing import Any, Callable


async def run_in_background(
    func: Callable, *args: Any, **kwargs: Any
) -> Any:
    """
    Run a function in the background to prevent blocking the main thread.  Functions will be executed using the default
    thread pool executor of the event loop.  By default, event loops use a ``concurrent.futures.ThreadPoolExecutor``,
    however it is recommended to override use ``concurrent.futures.ProcessPoolExecutor`` instead:

    ``asyncio.get_event_loop().set_default_executor(concurrent.futures.ProcessPoolExecutor)``

    Ref: https://github.com/encode/starlette/blob/master/starlette/concurrency.py#L21-L34
    """
    loop = asyncio.get_event_loop()
    func = partial(func, *args, **kwargs)
    return await loop.run_in_executor(None, func)


from dataclasses import dataclass
import math
from typing import Dict, Optional, Tuple, Union

import numpy as np

from .compression import Compression
from .constants import COMPRESSIONS, INTERLEAVE, SAMPLE_DTYPES
from .errors import TileNotFoundError
from .filesystems import Filesystem
from .tag import Tag
from .utils import run_in_background

@dataclass
class IFD:
    next_ifd_offset: int
    tag_count: int
    _file_reader: Filesystem

    @staticmethod
    def _is_masked(tiff_tags: Dict[str, Tag]) -> bool:
        """Check if an IFD is masked based on a dictionary of tiff tags"""
        # # https://www.awaresystems.be/imaging/tiff/tifftags/newsubfiletype.html
        # # https://gdal.org/drivers/raster/gtiff.html#internal-nodata-masks
        if "NewSubfileType" in tiff_tags:
            compression = tiff_tags['Compression'].value
            photo_interp = tiff_tags['PhotometricInterpretation'].value
            subfile_type = tiff_tags['NewSubfileType'].value
            if (subfile_type[0] == 1 or subfile_type[2] == 1) and photo_interp == 4 and compression == 8:
                return True
        return False

    @classmethod
    async def read(cls, file_reader: Filesystem) -> Union["ImageIFD", "MaskIFD"]:
        """Read the IFD"""
        ifd_start = file_reader.tell()
        tag_count = await file_reader.read(2, cast_to_int=True)
        tiff_tags = {}

        # Read tags
        for idx in range(tag_count):
            tag = await Tag.read(file_reader)
            if tag:
                tiff_tags[tag.name] = tag
        file_reader.seek(ifd_start + (12 * tag_count) + 2)
        next_ifd_offset = await file_reader.read(4, cast_to_int=True)

        # Check if mask
        if cls._is_masked(tiff_tags):
            return MaskIFD(next_ifd_offset, tag_count, file_reader, **tiff_tags)
        return ImageIFD(next_ifd_offset, tag_count, file_reader, **tiff_tags)

@dataclass
class RequiredTags:
    BitsPerSample: Tag
    Compression: Tag
    ImageHeight: Tag
    ImageWidth: Tag
    PhotometricInterpretation: Tag
    PlanarConfiguration: Tag
    SampleFormat: Tag
    SamplesPerPixel: Tag
    TileByteCounts: Tag
    TileHeight: Tag
    TileOffsets: Tag
    TileWidth: Tag

@dataclass
class OptionalTags:
    # TIFF standard tags
    NewSubfileType: Tag = None
    Predictor: Tag = None
    JPEGTables: Tag = None

    # GeoTiff
    GeoKeyDirectoryTag: Tag = None
    ModelPixelScaleTag: Tag = None
    ModelTiepointTag: Tag = None

    # GDAL private tags
    NoData: Tag = None


@dataclass
class ImageIFD(OptionalTags, Compression, RequiredTags, IFD):

    @property
    def compression(self) -> str:
        """Return the compression of the IFD"""
        return COMPRESSIONS[self.Compression.value]

    @property
    def bands(self) -> int:
        """Return the number of image bands"""
        return self.SamplesPerPixel.value

    @property
    def dtype(self) -> np.dtype:
        """Return the data type of the IFD"""
        if self.bands == 1:
            return np.dtype(
                SAMPLE_DTYPES[(self.SampleFormat.value, self.BitsPerSample.value)]
            )
        else:
            return np.dtype(
                SAMPLE_DTYPES[(self.SampleFormat.value[0], self.BitsPerSample.value[0])]
            )

    @property
    def nodata(self) -> Optional[int]:
        return int(self.NoData.value[0]) if self.NoData else None

    @property
    def interleave(self) -> str:
        """Return the interleave of the IFD"""
        return "band" if self.bands == 1 else INTERLEAVE[self.PlanarConfiguration.value]

    @property
    def is_full_resolution(self) -> bool:
        """Check if the IFD contains a full resolution image (not an overview)"""
        if not self.NewSubfileType:
            return True
        elif self.NewSubfileType.value[0] == 0:
            return False
        return True

    async def _get_tile(self, x: int, y: int) -> np.ndarray:
        """Read the requested tile from the IFD"""
        idx = (y * self.tile_count[0]) + x
        offset = self.TileOffsets[idx]
        byte_count = self.TileByteCounts[idx] - 1
        img_bytes = await self._file_reader.range_request(offset, byte_count)
        return await run_in_background(self._decompress, img_bytes)

    @property
    def tile_count(self) -> Tuple[int, int]:
        """Return the number of x/y tiles in the IFD"""
        return (
            math.ceil(self.ImageWidth.value / float(self.TileWidth.value)),
            math.ceil(self.ImageHeight.value / float(self.TileHeight.value)),
        )

    def __iter__(self):
        """Iterate through TIFF Tags"""
        for (k, v) in self.__dict__.items():
            if k not in ("next_ifd_offset", "tag_count", "_file_reader") and v:
                yield v


class MaskIFD(ImageIFD):

    async def _get_tile(self, x: int, y: int) -> np.ndarray:
        """Read the requested tile from the IFD"""
        idx = (y * self.tile_count[0]) + x
        offset = self.TileOffsets[idx]
        byte_count = self.TileByteCounts[idx] - 1
        img_bytes = await self._file_reader.range_request(offset, byte_count)
        return await run_in_background(self._decompress_mask, img_bytes)


import abc
import asyncio
from dataclasses import dataclass
import logging
import time
from typing import Callable
from urllib.parse import urlsplit

from aiocache import cached, Cache
import aiofiles
import aiohttp
import botocore.exceptions

from . import config

# https://github.com/developmentseed/rio-viz/blob/master/rio_viz/app.py#L33-L38
try:
    import aioboto3
    has_s3 = True
    if config.VERBOSE_LOGS:
        # Default to boto3 debug logs if verbose logging is enabled
        s3_log_level = config.LOG_LEVEL
    else:
        s3_log_level = logging.ERROR

    logging.getLogger("aiobotocore").setLevel(s3_log_level)
    logging.getLogger("botocore").setLevel(s3_log_level)
    logging.getLogger("aioboto3").setLevel(s3_log_level)
except ModuleNotFoundError:
    has_s3 = False


logger = logging.getLogger(__name__)
logger.setLevel(config.LOG_LEVEL)


def config_cache(fn: Callable) -> Callable:
    """
    Inject cache config params (https://aiocache.readthedocs.io/en/latest/decorators.html#aiocache.cached)
    """
    def wrap_function(*args, **kwargs):
        kwargs['cache_read'] = kwargs['cache_write'] = config.ENABLE_CACHE
        return fn(*args, **kwargs)
    return wrap_function


@dataclass
class Filesystem(abc.ABC):
    filepath: str

    def __post_init__(self):
        self.data: bytes = b""
        self._offset: int = 0
        self._endian: str = "<"
        self._total_bytes_requested: int = 0
        self._total_requests: int = 0
        self._header_size: int = config.INGESTED_BYTES_AT_OPEN
        self._requested_ranges = []

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        ...

    @classmethod
    def create_from_filepath(cls, filepath: str) -> "Filesystem":
        """Instantiate the appropriate filesystem based on filepath scheme"""
        splits = urlsplit(filepath)
        if splits.scheme in {"http", "https"}:
            return HttpFilesystem(filepath)
        elif splits.scheme == "s3":
            if not has_s3:
                raise NotImplementedError("Package must be built with [s3] extra to read from S3")
            return S3Filesystem(filepath)
        elif (not splits.scheme and not splits.netloc):
            return LocalFilesystem(filepath)
        raise NotImplemented("Unsupported file system")

    @config_cache
    @cached(
        cache=Cache.MEMORY,
        key_builder=lambda fn,*args,**kwargs: f"{args[0].filepath}-{args[1]}-{args[2]}"
    )
    async def range_request(self, start: int, offset: int) -> bytes:
        """
        Perform and cache a range request.
        """
        return await self._range_request(start, offset)

    @abc.abstractmethod
    async def _range_request(self, start: int, offset: int) -> bytes:
        """Perform a range request"""
        ...

    @abc.abstractmethod
    async def _close(self) -> None:
        """
        Close any resources created in ``__aexit__``, allows extending ``Filesystem`` context managers past their scope
        """
        ...


    async def read(self, offset: int, cast_to_int: bool = False):
        """
        Read from the current offset (self._offset) to the specified offset and optionall cast the result to int
        """
        if self._offset + offset > len(self.data):
            self.data += await self.range_request(len(self.data), config.INGESTED_BYTES_AT_OPEN)
        data = self.data[self._offset : self._offset + offset]
        self.incr(offset)
        order = "little" if self._endian == "<" else "big"
        return int.from_bytes(data, order) if cast_to_int else data

    def incr(self, offset: int) -> None:
        """Increment offset"""
        self._offset += offset

    def seek(self, offset: int) -> None:
        """Seek to the specified offset (setter for ``self._offset``)"""
        self._offset = offset

    def tell(self) -> int:
        """Return the current offset (getter for ``self._offset``)"""
        return self._offset


@dataclass
class HttpFilesystem(Filesystem):

    async def _range_request(self, start: int, offset: int) -> bytes:
        range_header = {"Range": f"bytes={start}-{start + offset}"}
        try:
            async with self.session.get(self.filepath, headers=range_header) as resp:
                resp.raise_for_status()
                data = await resp.content.read()
        except (aiohttp.ClientError, aiohttp.ClientResponseError) as e:
            await self._close()
            raise FileNotFoundError(f"File not found: {self.filepath}") from e
        return data

    async def _close(self) -> None:
        await self.session.close()

    async def __aenter__(self):
        trace_config = aiohttp.TraceConfig()
        trace_config.on_request_start.append((self._on_request_start))
        trace_config.on_request_end.append(self._on_request_end)
        self.session = aiohttp.ClientSession(trace_configs=[trace_config])
        return self

    async def _on_request_start(self, session, trace_config_ctx, params):
        trace_config_ctx.start = asyncio.get_event_loop().time()
        if config.VERBOSE_LOGS:
            debug_statement = (
                f"\n > {params.method} {params.url.path} HTTP/{session.version.major}.{session.version.minor}"
                f"\n   Host: {params.url.host}"
                f"\n   Range: {params.headers['Range']}"
            )
        else:
            debug_statement = f" STARTING REQUEST: {params.method} {params.url}"
        logger.debug(debug_statement)

    async def _on_request_end(self, session, trace_config_ctx, params):
        if params.response.status < 400:
            elapsed = round(asyncio.get_event_loop().time() - trace_config_ctx.start, 3)
            content_range = params.response.headers['Content-Range']
            self._total_bytes_requested += int(params.response.headers["Content-Length"])
            self._total_requests += 1
            self._requested_ranges.append(tuple([int(v) for v in content_range.split(' ')[-1].split('/')[0].split('-')]))
            if config.VERBOSE_LOGS:
                debug_statement = [f"\n < HTTP/{session.version.major}.{session.version.minor}"]
                debug_statement += [f"\n < {k}: {v}" for (k, v) in params.response.headers.items()]
                debug_statement.append(f"\n < Duration: {elapsed}")
            else:
                debug_statement = f" FINISHED REQUEST in {elapsed} seconds: <STATUS {params.response.status}> ({content_range})"
            logger.debug("".join(debug_statement))


@dataclass
class LocalFilesystem(Filesystem):

    async def _range_request(self, start: int, offset: int) -> bytes:
        begin = time.time()
        await self.file.seek(start)
        data = await self.file.read(offset+1)
        elapsed = time.time() - begin
        self._total_bytes_requested += (offset - start + 1)
        self._total_requests += 1
        self._requested_ranges.append((start, start+offset))
        logger.debug(f" FINISHED REQUEST in {elapsed} seconds: <STATUS 206> ({start}-{start+offset})")
        return data

    async def _close(self) -> None:
        await self.file.close()

    async def __aenter__(self):
        self.file = await aiofiles.open(self.filepath, 'rb')
        return self


@dataclass
class S3Filesystem(Filesystem):

    async def _range_request(self, start: int, offset: int) -> bytes:
        begin = time.time()
        try:
            req = await self.object.get(Range=f'bytes={start}-{start+offset}')
        except botocore.exceptions.ClientError as e:
            await self._close()
            raise FileNotFoundError(f"File not found: {self.filepath}") from e
        elapsed = time.time() - begin
        content_range = req['ResponseMetadata']['HTTPHeaders']['content-range']
        if not config.VERBOSE_LOGS:
            status = req['ResponseMetadata']['HTTPStatusCode']
            logger.debug(f" FINISHED REQUEST in {elapsed} seconds: <STATUS {status}> ({content_range})")
        self._total_bytes_requested += int(req['ResponseMetadata']['HTTPHeaders']['content-length'])
        self._total_requests += 1
        self._requested_ranges.append(tuple([int(v) for v in content_range.split(' ')[-1].split('/')[0].split('-')]))
        data = await req['Body'].read()
        return data

    async def _close(self) -> None:
        await self.resource.__aexit__('', '', '')

    async def __aenter__(self):
        splits = urlsplit(self.filepath)
        self.resource = await aioboto3.resource('s3').__aenter__()
        self.object = await self.resource.Object(splits.netloc, splits.path[1:])
        return self


import asyncio
from dataclasses import dataclass, field
import logging
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
from urllib.parse import urljoin
import uuid

import affine
from PIL import Image
import numpy as np

from . import config
from .constants import PHOTOMETRIC
from .errors import InvalidTiffError, TileNotFoundError
from .filesystems import Filesystem
from .ifd import IFD, ImageIFD, MaskIFD
from .partial_reads import PartialReadInterface
from .utils import run_in_background

logger = logging.getLogger(__name__)
logger.setLevel(config.LOG_LEVEL)


@dataclass
class COGReader(PartialReadInterface):
    filepath: str
    ifds: Optional[List[ImageIFD]] = field(default_factory=lambda: [])
    mask_ifds: Optional[List[MaskIFD]] = field(default_factory=lambda: [])

    _version: Optional[int] = 42
    _big_tiff: Optional[bool] = False


    async def __aenter__(self):
        """Open the image and read the header"""
        async with Filesystem.create_from_filepath(self.filepath) as file_reader:
            self._file_reader = file_reader
            if (await file_reader.read(2)) == b"MM":
                file_reader._endian = ">"
            version = await file_reader.read(2, cast_to_int=True)
            if version == 42:
                first_ifd = await file_reader.read(4, cast_to_int=True)
                file_reader.seek(first_ifd)
                await self._read_header()
            elif version == 43:
                raise NotImplementedError("BigTiff is not yet supported")
            else:
                raise InvalidTiffError("Not a valid TIFF")
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._file_reader._close()

    def __iter__(self):
        """Iterate through image IFDs"""
        for ifd in self.ifds:
            yield ifd

    @property
    def profile(self) -> Dict[str, Any]:
        """Return a rasterio-style image profile"""
        ifd = self.ifds[0]
        return {
            "driver": "GTiff",
            "width": ifd.ImageWidth.value,
            "height": ifd.ImageHeight.value,
            "count": ifd.bands,
            "dtype": str(ifd.dtype),
            "transform": self.geotransform(),
            "blockxsize": ifd.TileWidth.value,
            "blockysize": ifd.TileHeight.value,
            "compress": ifd.compression,
            "interleave": ifd.interleave,
            "crs": f"EPSG:{self.epsg}",
            "nodata": ifd.nodata,
            "tiled": True,
            "photometric": PHOTOMETRIC[ifd.PhotometricInterpretation.value],
        }

    @property
    def epsg(self) -> int:
        """Return the EPSG code representing the crs of the image"""
        ifd = self.ifds[0]
        for idx in range(0, len(ifd.GeoKeyDirectoryTag), 4):
            # 2048 is geographic crs
            # 3072 is projected crs
            if ifd.GeoKeyDirectoryTag[idx] in (2048, 3072):
                return ifd.GeoKeyDirectoryTag[idx + 3]

    @property
    def bounds(self) -> Tuple[float, float, float, float]:
        """Return the bounds of the image in native crs"""
        gt = self.geotransform()
        tlx = gt.c
        tly = gt.f
        brx = tlx + (gt.a * self.ifds[0].ImageWidth.value)
        bry = tly + (gt.e * self.ifds[0].ImageHeight.value)
        return (tlx, bry, brx, tly)

    @property
    def overviews(self) -> List[int]:
        """Return decimation factor for each overview (2**zoom)"""
        return [2 ** (ifd + 1) for ifd in range(len(self.ifds) - 1)]

    @property
    def requests(self) -> Dict[str, Union[int, List[Tuple[int]]]]:
        """Return statistics about http requests made during context lifecycle"""
        return {
            'count': self._file_reader._total_requests,
            'byte_count': self._file_reader._total_bytes_requested,
            'ranges': self._file_reader._requested_ranges,
            'header_size': self._file_reader._header_size
        }

    @property
    def is_masked(self) -> bool:
        """Check if the image has an internal mask"""
        return True if self.mask_ifds else False

    @property
    def nodata(self) -> Optional[int]:
        return self.ifds[0].nodata

    async def _read_header(self) -> None:
        """Internal method to read image header and parse into IFDs and Tags"""
        next_ifd_offset = 1
        while next_ifd_offset != 0:
            ifd = await IFD.read(self._file_reader)
            logger.debug(f" Opened {ifd.ImageHeight.value}x{ifd.ImageWidth.value} overview")
            next_ifd_offset = ifd.next_ifd_offset
            self._file_reader.seek(next_ifd_offset)

            if isinstance(ifd, MaskIFD):
                self.mask_ifds.append(ifd)
            else:
                self.ifds.append(ifd)


    def geotransform(self, ovr_level: int = 0) -> affine.Affine:
        """Return the geotransform of the image at a specific overview level (defaults to native resolution)"""
        # Calculate overview for source image
        gt = affine.Affine(
            self.ifds[0].ModelPixelScaleTag[0],
            0.0,
            self.ifds[0].ModelTiepointTag[3],
            0.0,
            -self.ifds[0].ModelPixelScaleTag[1],
            self.ifds[0].ModelTiepointTag[4],
        )
        # Decimate the geotransform if an overview is requested
        if ovr_level > 0:
            bounds = self.bounds
            ifd = self.ifds[ovr_level]
            gt = affine.Affine.translation(bounds[0], bounds[3]) * affine.Affine.scale(
                (bounds[2] - bounds[0]) / ifd.ImageWidth.value,
                (bounds[1] - bounds[3]) / ifd.ImageHeight.value,
            )
        return gt


    async def get_tile(self, x: int, y: int, z: int) -> np.ndarray:

        """
        Request an internal image tile at the specified row (x), column (y), and overview (z).  Based on COGDumper:
        https://github.com/mapbox/COGDumper/blob/master/cogdumper/cog_tiles.py#L337-L365
        """
        futures = []
        if z > len(self.ifds):
            raise TileNotFoundError(f"Overview {z} does not exist.")
        ifd = self.ifds[z]
        xmax, ymax = ifd.tile_count

        # Return an empty array if tile is outside bounds of image
        if x < 0 or y < 0 or x >= xmax or y >= ymax:
            if not config.BOUNDLESS_READ:
                raise TileNotFoundError(f"Internal tile {z}/{x}/{y} does not exist")
            tile = np.full(
                (ifd.bands, ifd.TileHeight.value, ifd.TileWidth.value),
                fill_value=config.BOUNDLESS_READ_FILL_VALUE
            )
            return tile

        # Request the tile
        futures.append(
            asyncio.create_task(ifd._get_tile(x, y))
        )

        # Request the mask
        if self.is_masked:
            mask_ifd = self.mask_ifds[z]
            futures.append(
                asyncio.create_task(mask_ifd._get_tile(x, y))
            )

        tile = await asyncio.gather(*futures)

        # Prioritize internal mask over nodata
        if self.is_masked:
            # Apply mask
            tile[1] = np.invert(np.broadcast_to(tile[1], tile[0].shape))
            return np.ma.masked_array(*tile)
        # Explicitly check for None because nodata is often 0
        if ifd.nodata is not None:
            return np.ma.masked_where(tile[0] == ifd.nodata, tile[0])
        return tile[0]

    async def read(
        self,
        bounds: Tuple[float, float, float, float],
        shape: Tuple[int, int],
        resample_method: int = Image.NEAREST,
    ) -> Union[np.ndarray, np.ma.masked_array]:
        """
        Perform a partial read.  All pixels within the specified bounding box are read from the image and the array is
        resampled to match the desired shape.
        """
        # Determine which tiles intersect the request bounds
        ovr_level = self._get_overview_level(bounds, shape[1], shape[0])
        ifd = self.ifds[ovr_level]
        img_tiles = self._calculate_image_tiles(
            bounds,
            tile_width=ifd.TileWidth.value,
            tile_height=ifd.TileHeight.value,
            band_count=ifd.bands,
            ovr_level=ovr_level,
            dtype=ifd.dtype
        )

        if not self._intersect_bounds(bounds, self.bounds):
            raise TileNotFoundError("Partial read is outside bounds of the image")

        # Request those tiles
        if config.HTTP_MERGE_CONSECUTIVE_RANGES:
            img_arr = await self._request_merged_tiles(img_tiles)
        else:
            img_arr = await self._request_tiles(img_tiles)

        # Postprocess the array (clip to bounds and resize to requested shape)
        postprocessed = await run_in_background(
            self._postprocess,
            arr=img_arr,
            img_tiles=img_tiles,
            out_shape=shape,
            resample_method=resample_method
        )

        return postprocessed

    def create_tile_matrix_set(self, identifier: str = None) -> Dict[str, Any]:
        """Create an OGC TileMatrixSet where each TileMatrix corresponds to an overview"""
        matrices = []
        for idx, ifd in enumerate(self.ifds):
            gt = self.geotransform(idx)
            matrix = {
                "identifier": str(len(self.ifds) - idx - 1),
                "topLeftCorner": [gt.c, gt.f],
                "tileWidth": ifd.TileWidth.value,
                "tileHeight": ifd.TileHeight.value,
                "matrixWidth": ifd.tile_count[0],
                "matrixHeight": ifd.tile_count[1],
                "scaleDenominator": gt.a / 0.28e-3,
            }
            matrices.append(matrix)

        tms = {
            "title": f"Tile matrix for {self.filepath}",
            "identifier": identifier or str(uuid.uuid4()),
            "supportedCRS": urljoin(f"http://www.opengis.net", f"/def/crs/EPSG/0/{self.epsg}"),
            "tileMatrix": list(reversed(matrices))
        }
        return tms
